@@ -1,11 +1,11 @@
 ; -*- Mode: Scheme; Syntax: Scheme; Package: Scheme; -*-
-; Copyright (c) 1993-2005 by Richard Kelsey and Jonathan Rees. See file COPYING.
+; Copyright (c) 1993-2007 by Richard Kelsey and Jonathan Rees. See file COPYING.
 
 ; This is file arch.scm.
 
 ;;;; Architecture description
 
-(define architecture-version "Vanilla 27")
+(define architecture-version "Vanilla 32")
 
 ; Things that the VM and the runtime system both need to know.
 
@@ -20,12 +20,16 @@
 ; (op . args)
 ; OP may be a name or a list of names
 ; ARGS are
-;  nargs        - a byte
-;  byte         - a byte
-;  two-bytes    - two bytes
+
+;  byte                   - A byte used for some random purpose
+;  two-bytes              - ditto, but two bytes
+;  nargs                  - number of arguments on the stack
+;  two-byte-nargs         - ditto, but two bytes
+;  stack-index            - A one-byte index into the stack
+;  two-byte-stack-index   - ditto, but two bytes
+;  index                  - A index into some stob
+;  two-byte-index         - ditto, but two bytes
 ;  literal      - a one-byte value
-;  index        - a two byte index into the current template or environment
-;  small-index  - a one byte index into the current template or environment
 ;  offset       - two bytes giving an offset into the current instruction stream
 ;  offset-      - same thing, negative offset
 ;  stob         - a byte specifying a type for a stored object
@@ -68,12 +72,21 @@
   (push+integer-literal literal) ; preceded by a push
   (integer-literal+push literal) ; followed by a push
 
-  (global      two-bytes index)   ; first is template, second within template
-  (set-global! two-bytes index 1) ; first is template, second within template
+  (global      two-byte-stack-index two-byte-index)   ; first is template, second within template
+  (set-global! two-byte-stack-index two-byte-index 1) ; first is template, second within template
 
   (make-flat-env  env-data)      ; make new environment from env-data
   (make-big-flat-env big-env-data) ; same, but with two-byte size and offsets
 
+  ; the following four emitted from the byte-code optimizer, for the
+  ; benefit of the native-code compiler
+  (env-set! stack-index index 1)             ; set environment slot
+  (big-env-set! two-byte-stack-index two-byte-index 1)
+  (template-ref stack-index index) ; same thing as stack-indirect
+  (big-template-ref two-byte-stack-index two-byte-index) ; same thing as stack-indirect
+
+  (make-flat-closure two-bytes)  ; create flat closure
+
   (push 1)		         ; push *val* onto stack
   (push-false)			 ; a common combination
   (pop)			         ; pop top of stack into *val*
@@ -81,17 +94,17 @@
 				 ; leaving *val* unchanged
   (push-n	  two-bytes)     ; allocate space for N values on stack
 				 ; leaving *val* unchanged
-  (stack-ref      byte)	         ; index'th element of stack into *val*
-  (push+stack-ref byte)	         ; preceded by a push
-  (stack-ref+push byte)	         ; followed by a push
-  (big-stack-ref  two-bytes)
-  (stack-set!     byte 1)        ; *val* to index'th element of stack
-  (big-stack-set! two-bytes 1)
-  (stack-indirect byte byte)     ; first is index into stack, second is index
+  (stack-ref      stack-index)   ; index'th element of stack into *val*
+  (push+stack-ref stack-index)   ; preceded by a push
+  (stack-ref+push stack-index)   ; followed by a push
+  (big-stack-ref  two-byte-stack-index)
+  (stack-set!     stack-index 1) ; *val* to index'th element of stack
+  (big-stack-set! two-byte-stack-index 1)
+  (stack-indirect stack-index index)     ; first is index into stack, second is index
 			  	 ; into what you find there
-  (push+stack-indirect byte byte) ; preceded by a push
-  (stack-indirect+push byte byte) ; followed by a push
-  (big-stack-indirect two-bytes two-bytes)
+  (push+stack-indirect stack-index index) ; preceded by a push
+  (stack-indirect+push stack-index index) ; followed by a push
+  (big-stack-indirect two-byte-stack-index two-byte-index)
 
   (stack-shuffle! moves-data)	 ; shuffle stack elements around
   (big-stack-shuffle! big-moves-data)	 ; shuffle stack elements around with two-byte offsets
@@ -104,9 +117,9 @@
   (call     offset nargs 1 +)    ; last argument is the procedure to call,
 				 ; offset is to return pointer
   (tail-call nargs 1 +)          ; same, no return pointer, moves arguments
-  (big-call offset two-bytes 1 +) ; ditto, nargs counts are two bytes
-  (poll offset)			 ; offset is for continuation-pc if interrupt
-  (apply offset two-bytes 2 +)   ; last argument is procedure to call, second to
+  (big-call offset two-byte-nargs 1 +) ; ditto, nargs counts are two bytes
+  (poll)
+  (apply offset two-byte-nargs 2 +)   ; last argument is procedure to call, second to
 				 ; last is a list of additional arguments, next
                                  ; two bytes are the number of stack arguments
   (closed-apply 2 +)		 ; arguments are as for Scheme's APPLY, with
@@ -120,9 +133,9 @@
   (closed-values +)		 ; values are on stack, count is pushed on stack
 
   ;; Six different ways to jump
-  (goto-template index)		 ; jump to another template (*EXP*)
+  (goto-template two-byte-index)		 ; jump to another template (*EXP*)
 				 ; does not poll for interrupts
-  (call-template offset index index nargs)
+  (call-template offset two-byte-stack-index two-byte-index nargs)
 				 ; call a template instead of a procedure
 				 ; nargs is needed for interrupt handling
   (jump-if-false offset 1)	 ; boolean in *val*
@@ -162,7 +175,7 @@
   (arithmetic-shift 2)
   (char? 1)
   ((char=? char<?) 2)
-  ((char->ascii ascii->char) 1)
+  ((char->scalar-value scalar-value->char scalar-value?) 1)
   (eof-object? 1)
 
   ;; Data manipulation
@@ -171,8 +184,8 @@
 
   (make-stored-object byte stob)
   (closed-make-stored-object stob)	; size pushed on stack
-  (stored-object-ref  stob byte 1)	; byte is the offset
-  (stored-object-set! stob byte byte 2)	; byte0 is the offset, byte1 = 0
+  (stored-object-ref  stob index 1)
+  (stored-object-set! stob index byte 2)	; byte = 0
   					; means not to log in the proposal
 
   (make-vector-object stob 2)			; size + init
@@ -180,6 +193,8 @@
   (stored-object-indexed-ref  stob byte 2)	; vector + offset
   (stored-object-indexed-set! stob byte 3)	; vector + offset + value
 
+  (make-double)
+
   (make-byte-vector 2)
   (byte-vector-length 1)
   (byte-vector-ref 2)
@@ -189,6 +204,7 @@
   (string-length 1)
   (string-ref 2)
   (string-set! 3)
+  (copy-string-chars! 5)
 
   (intern 1)
                      
@@ -197,10 +213,11 @@
   ((immutable? make-immutable!) 1)
 
   ;; channels (unbuffered, non-blocking I/O)
-  (open-channel 3)
+  (open-channel 4)
   (close-channel 1)
   (channel-maybe-read 5)
   (channel-maybe-write 4)
+  (channel-parameter 1)
   (channel-ready? 1)
   (channel-abort 1)             ; stop channel operation
   (open-channels-list)		; return a list of the open channels
@@ -209,7 +226,7 @@
   (current-proposal)
   (set-current-proposal! 1)
   (maybe-commit)
-  (stored-object-logging-ref stob byte 1)
+  (stored-object-logging-ref stob index 1)
   (copy-bytes! byte 5)		; byte = 0 -> don't log
   (byte-vector-logging-ref 2)
   (byte-vector-logging-set! 3)
@@ -235,7 +252,9 @@
   (return-from-native-exception 1)
   (set-interrupt-handlers! 1)
   (set-enabled-interrupts! 1)
-  (return-from-interrupt)
+  (resume-interrupted-opcode-to-byte-code)
+  (resume-interrupted-call-to-native-code)
+  (resume-native-poll)
   (schedule-interrupt 1)
   (wait 2)                      ; do nothing until something happens
   (call-external-value 1 +)
@@ -243,6 +262,7 @@
   (undefine-shared-binding 2)
   (find-undefined-imported-bindings)
   (time 2)
+  (system-parameter 1)
   (vm-extension 2)		; access to extensions of the virtual machine
   (return-from-callback 2)	; return from an callback
 
@@ -252,15 +272,24 @@
   (assq 2)
   (unassigned-check 1)
   ; If the byte = 0 then do not log in the current proposal
-  (checked-record-ref byte 3)
-  (checked-record-set! byte 4)
+  (checked-record-ref index 3)
+  (checked-record-set! index 4)
+
+  (encode-char 5)
+  (encode-char! 5)
+
+  (decode-char 4)
+  (decode-char! 4)
 
   ;; ports (buffered I/O) - these are all unnecessary
   ;; byte = 0 -> port is supplied
   ;;      = 1 -> get port from dynamic environment
+  ((read-byte peek-byte) byte 1 0)
+  (write-byte byte 2 1)
+
   ((read-char peek-char) byte 1 0)
   (write-char byte 2 1)
-  
+
   (os-error-message 1)
 
   ;; For writing informative messages when debugging
@@ -309,7 +338,7 @@
    illegal-exception-return
    ))
 
-; Used by (READ-CHAR) and (WRITE-CHAR) to get the appropriate ports from
+; Used by (READ-BYTE) and (WRITE-BYTE) to get the appropriate ports from
 ; the fluid environment.
 
 (define-enumeration current-port-marker
@@ -410,6 +439,13 @@
 (define continuation-code-index 1)
 (define continuation-cont-index 2)
 
+; Offsets in the CONT-DATA instruction
+					; -1 -2        frame size
+(define gc-mask-size-offset -3)		; -3           gc mask size
+					; -4 -5        offset
+                                        ; -6 -7        template
+(define gc-mask-offset      -8)         ; -8 ...       mask (low bytes first)
+
 ; The number of additional values that the VM adds to exception continuations.
 (define exception-continuation-cells 5)
 
@@ -462,6 +498,25 @@
    open-for-output
    ))
 
+; Parameters that configure a channel
+
+(define-enumeration channel-parameter-option
+  (buffer-size crlf?))
+
+; Built-in text encodings
+
+(define-enumeration text-encoding-option
+  (us-ascii
+   latin-1
+   utf-8
+   utf-16le utf-16be
+   utf-32le utf-32be))
+
+; Options for op/system-parameter
+
+(define-enumeration system-parameter-option
+  (os-string-encoding))
+
 (define-enumeration stob
   (;; D-vector types (traced by GC)
    pair
@@ -480,7 +535,6 @@
    weak-pointer
    shared-binding
    unused-d-header1
-   unused-d-header2
 
    ;; B-vector types (not traced by GC)
    string        ; = least b-vector type
@@ -516,12 +570,19 @@
       (shared-binding-ref shared-binding-set!))
     (port port? make-port
       (port-handler)
+      ;; either an integer from the TEXT-ENCODING-OPTION for encodings
+      ;; handled by the VM, or a :TEXT-CODEC object for things handled
+      ;; purely by the RTS
+      (port-text-codec-spec set-port-text-codec-spec!)
+      (port-crlf?   set-port-crlf?!)
       (port-status  set-port-status!)
       (port-lock    set-port-lock!)		; used for buffer timestamps
       (port-data    set-port-data!)
       (port-buffer  set-port-buffer!)
       (port-index   set-port-index!)
       (port-limit   set-port-limit!)
+      ;; for CR/LF handling
+      (port-pending-cr?  set-port-pending-cr?!)
       (port-pending-eof? set-port-pending-eof?!))
     (channel channel? #f
       (channel-status)
