@@ -1,5 +1,5 @@
 ; -*- Mode: Scheme; Syntax: Scheme; Package: Scheme; -*-
-; Copyright (c) 1993-2005 by Richard Kelsey and Jonathan Rees. See file COPYING.
+; Copyright (c) 1993-2007 by Richard Kelsey and Jonathan Rees. See file COPYING.
 
 ;Need to fix the byte-code compiler to make jump etc. offsets from the
 ;beginning of the instruction.  (Later: why?)
@@ -35,7 +35,9 @@
 (define *interrupted-template*) ; template in place when the most recent
 				; interrupt occured - for profiling
 
-(define *interrupt-return-code*)	; used to return from interrupts
+(define *interrupted-byte-opcode-return-code*)	; used to return from interrupts
+(define *interrupted-native-call-return-code*)	; used to return from interrupts
+(define *native-poll-return-code*)              ; used to return from interrupts
 (define *exception-return-code*)	; used to mark exception continuations
 (define *native-exception-return-code*)	; used to mark native exception continuations
 (define *call-with-values-return-code*)	; for call-with-values opcode
@@ -59,7 +61,9 @@
 
 (define (clear-registers)
   (reset-stack-pointer false)
-  (set-code-pointer! *interrupt-return-code* 0)
+  (set-code-pointer! *interrupted-byte-opcode-return-code* 0)
+  (set-code-pointer! *interrupted-native-call-return-code* 0)
+  (set-code-pointer! *native-poll-return-code* 0)
   (set! *last-code-pointer-resumed* *code-pointer*)
   (set! *val*                unspecific-value)
   (set! *current-thread*     null)
@@ -96,7 +100,12 @@
 
     (set! *val*                   (s48-trace-value *val*))
     (set! *current-thread*        (s48-trace-value *current-thread*))
-    (set! *interrupt-return-code* (s48-trace-value *interrupt-return-code*))
+    (set! *interrupted-byte-opcode-return-code*
+          (s48-trace-value *interrupted-byte-opcode-return-code*))
+    (set! *interrupted-native-call-return-code*
+          (s48-trace-value *interrupted-native-call-return-code*))
+    (set! *native-poll-return-code*
+          (s48-trace-value *native-poll-return-code*))
     (set! *exception-return-code* (s48-trace-value *exception-return-code*))
     (set! *native-exception-return-code* (s48-trace-value *native-exception-return-code*))
     (set! *call-with-values-return-code*
@@ -115,8 +124,7 @@
     (trace-finalizer-alist!)
     
     ; These could be moved to the appropriate modules.
-    (trace-io s48-trace-value)
-    (trace-channel-names s48-trace-value)))
+    (trace-io s48-trace-value)))
 
 (add-post-gc-cleanup!
   (lambda ()
@@ -124,6 +132,7 @@
     (set! *last-code-pointer-resumed* *code-pointer*)
     (partition-finalizer-alist!)
     (close-untraced-channels! s48-extant? s48-trace-value)
+    (trace-channel-names s48-trace-value)
     (note-interrupt! (enum interrupt post-gc))))
 
 ;----------------
@@ -218,7 +227,10 @@
       (let ((code (code-pointer->code *last-code-pointer-resumed*)))
 	(if (within-code? *code-pointer* code)
 	    code
-	    (error "VM error: unable to locate current code vector")))))
+	    (error "VM error: unable to locate current code vector"
+                   (address->integer *code-pointer*)
+                   *last-code-called*
+                   (address->integer *last-code-pointer-resumed*))))))
 
 (define (code-pointer->code code-pointer)
   (let ((pc (fetch-two-bytes (address- code-pointer 5))))
@@ -241,24 +253,40 @@
 ; is in the continuation just below the saved registers.
 
 (define (initialize-interpreter+gc)          ;Used only at startup
-  (let ((key (ensure-space (* 4 return-code-size))))
-    (set! *interrupt-return-code*
+  (let ((key (ensure-space (* 6 return-code-size))))
+    (set! *interrupted-byte-opcode-return-code*
 	  (make-return-code ignore-values-protocol
-			    (enum op return-from-interrupt)
+                            #xffff              ; dummy template offset
+			    (enum op resume-interrupted-opcode-to-byte-code)
+			    #xFFFF		; escape value
+			    key))
+    (set! *interrupted-native-call-return-code*
+	  (make-return-code ignore-values-protocol
+                            #xffff              ; dummy template offset
+			    (enum op resume-interrupted-call-to-native-code)
+			    #xFFFF		; escape value
+			    key))
+    (set! *native-poll-return-code*
+	  (make-return-code ignore-values-protocol
+                            #xffff              ; dummy template offset
+			    (enum op resume-native-poll)
 			    #xFFFF		; escape value
 			    key))
     (set! *exception-return-code*
 	  (make-return-code 1	; want exactly one return value
+                            #xffff              ; dummy template offset
 			    (enum op return-from-exception)
 			    #xFFFF		; escape value
 			    key))
     (set! *native-exception-return-code*
 	  (make-return-code 1	; want exactly one return value
+                            #xffff              ; dummy template offset
 			    (enum op return-from-native-exception)
 			    #xFFFF		; escape value
 			    key))
     (set! *call-with-values-return-code*
 	  (make-return-code call-with-values-protocol
+                            #xffff              ; dummy template offset
 			    0			; opcode is not used
 			    1
 			    key))))
@@ -483,6 +511,7 @@
     (set! *val* (vm-vector-ref handlers opcode))
     (if (not (closure? *val*))
 	(lose "exception handler is not a closure"))
+    ;; We add 2, one for the opcode, one for the exception itself
     (goto call-exception-handler (+ nargs 2) opcode)))
 
 ;----------------
@@ -496,6 +525,19 @@
 		      (code-byte 1))
 	2))
 
+; same as stack-indirect, but serves as annotation for native-code
+; compiler
+(define-opcode template-ref
+  (goto continue-with-value
+	(d-vector-ref (stack-ref (code-byte 0))
+		      (code-byte 1))
+	2))
+(define-opcode big-template-ref
+  (goto continue-with-value
+	(d-vector-ref (stack-ref (code-offset 0))
+		      (code-offset 2))
+	4))
+
 (define-opcode push+stack-indirect
   (push *val*)
   (goto continue-with-value
